# -*- coding: utf-8 -*-
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang import error, b
from golang import errors, _errors
from golang.golang_test import import_pyx_tests
from golang._errors_test import pyerror_mkchain as error_mkchain
from pytest import raises

import_pyx_tests("golang._errors_test")


# assertEeq asserts that `e1 == e2`, `not (e1 != e2)`, `hash(e1) == hash(e2)`,
# and symmetrically.
def assertEeq(e1, e2):
    assert e1 == e2
    assert e2 == e1
    assert not (e1 != e2)
    assert not (e2 != e1)
    assert hash(e1) == hash(e2)
    assert hash(e2) == hash(e1)

# assertEne asserts that `e1 != e2`, `not (e1 == e2)` and symmetrically.
def assertEne(e1, e2):
    assert e1 != e2
    assert e2 != e1
    assert not (e1 == e2)
    assert not (e2 == e1)

    # cannot - generally speaking there could be hash collisions
    #assert hash(e1) != hash(e2)


# EError is custom error class that inherits from error.
class EError(error):
    def __init__(myerr, op, ret):
        myerr.op  = op
        myerr.ret = ret

    def Error(myerr): return "my %s: %s" % (myerr.op, myerr.ret)
    # no .Unwrap()

    # NOTE error provides good __eq__ and __hash__ out of the box.


# EErrorWrap is custom error class that inherits from error and provides .Unwrap()
class EErrorWrap(error):
    def __init__(myw, text, err):
        myw.text = text
        myw.err  = err

    def Error(myw):  return "%s: %s" % (myw.text, myw.err)
    def Unwrap(myw): return myw.err

# XWrap is custom error class that provides .Unwrap(), but does not inherit
# from error, nor provides .Error().
class XWrap(BaseException): # NOTE does not inherit from error
    def __init__(xw, text, err):
        xw.text = text
        xw.err  = err

    def Unwrap(xw):  return xw.err
    def __str__(xw): return "%s: %s" % (xw.text, xw.err)
    # no .Error()

    def __eq__(a, b):
        return (type(a) is type(b)) and \
                (a.text == b.text) and (a.err == b.err)


# XExc is custom error class that does not inherit from error, nor provides
# .Error() nor .Unwrap().
class XExc(Exception): # NOTE does not inherit from error
    def __init__(xerr, text):
        xerr.text = text
    def __str__(xerr):  return xerr.text
    def __repr__(xerr): return 'XExc("%s")' % xerr.text
    # no .Error()
    # no .Unwrap()

    def __eq__(a, b):
        return (type(a) is type(b)) and \
                (a.text == b.text)


# Unwrap1(e) is approximate for `errors.Unwrap(e)` in Go.
def Unwrap1(e):
    wit = _errors._pyUnwrapIter(e)
    return next(wit, None)


# test for golang.error class.
def test_error():
    assert error_mkchain([]) is None

    e = error_mkchain(["abc"])
    assert type(e)    is error
    assert e.Error()  == "abc"
    assert str(e)     == "abc"
    assert repr(e).endswith(' error="abc">')
    assert e.Unwrap() is None
    assertEeq(e, e)

    e1 = e
    e = error_mkchain(["привет", "abc"])
    assert type(e)    is error
    assert e.Error()  == "привет: abc"
    assert str(e)     == "привет: abc"
    assert repr(e).endswith(' error="привет: abc">')
    e1_ = e.Unwrap()
    assert type(e1_) is type(e1)
    assertEeq(e1_, e1)
    assert e1_.Unwrap() is None
    assertEeq(e, e)
    assertEne(e, e1)

    e2 = e

    # create an error from py via error() call
    with raises(ValueError): error()
    with raises(ValueError): error("hello", "world")
    for x in ["hello мир", u"hello мир", b("hello мир")]:
        e = error(x)
        assert type(e)    is error
        assert e.Error()  == "hello мир"
        assert str(e)     == "hello мир"
        assert repr(e).endswith(' error="hello мир">')
        assert e.Unwrap() is None
        assertEeq(e, e)

    # create an error from py via error subclass
    class EErr(error): pass
    m = EErr("abc")
    assertEeq(m, m)
    assertEne(m, error("abc"))  # EErr("abc") != error("abc")

    epy = EError("load", 3)
    assert type(epy)    is EError
    assert epy.Error()  == "my load: 3"
    assert str(epy)     == "my load: 3"
    assert repr(epy)    == "golang.errors_test.EError('load', 3)"
    assert epy.Unwrap() is None
    assertEeq(epy, epy)
    assertEeq(epy, EError("load", 3))
    assertEne(epy, EError("load", 4))

    wpy = EErrorWrap("mywrap", epy)
    assert type(wpy)    is EErrorWrap
    assert wpy.Error()  == "mywrap: my load: 3"
    assert str(wpy)     == "mywrap: my load: 3"
    assert repr(wpy)    == "golang.errors_test.EErrorWrap('mywrap', golang.errors_test.EError('load', 3))"
    assert wpy.Unwrap() is epy
    assert epy.Unwrap() is None

    epy = RuntimeError("zzz")
    wpy = EErrorWrap("qqq", epy)
    assert type(wpy) is EErrorWrap
    assert wpy.Error()  == "qqq: zzz"
    assert str(wpy)     == "qqq: zzz"
    assert repr(wpy)    == "golang.errors_test.EErrorWrap('qqq', %r)" % epy
    assert wpy.Unwrap() is epy
    with raises(AttributeError): epy.Unwrap


def test_new():
    E = errors.New

    for x in ["мир", u"мир", b("мир")]:
        err = E(x)
        assert type(err) is error
        assertEeq(err, E("мир"))
        assertEne(err, E("def"))

        assertEeq(err, error("мир"))
        assertEeq(err, error(u"мир"))
        assertEeq(err, error(b("мир")))

    with raises(TypeError):
        E(1)


# verify Unwrap for simple linear cases.
def test_unwrap():
    E  = errors.New
    Ec = error_mkchain

    # err must be exception or None
    assert Unwrap1(None) is None
    assert Unwrap1(BaseException()) is None
    with raises(TypeError): Unwrap1(1)
    with raises(TypeError): Unwrap1(object())

    ewrap = Ec(["abc", "def", "zzz"])
    e1 = Unwrap1(ewrap)
    assertEeq(e1, Ec(["def", "zzz"]))
    e2 = Unwrap1(e1)
    assertEeq(e2, E("zzz"))
    assert Unwrap1(e2) is None

    # Python-level error class that define .Wrap()
    e = EError("try", "fail")
    w = EErrorWrap("topic", e)
    e1 = Unwrap1(w)
    assert e1 is e
    assert Unwrap1(e1) is None

    # same, but wrapped is !error
    e = RuntimeError("zzz")
    w = EErrorWrap("qqq", e)
    e1 = Unwrap1(w)
    assert e1 is e
    assert Unwrap1(e1) is None


# verify Is for simple linear cases.
def test_is():
    E = errors.New
    Ec = error_mkchain

    assert errors.Is(None,    None)     == True
    assert errors.Is(E("a"),  None)     == False
    assert errors.Is(None,    E("b"))   == False

    # don't accept !error err
    assert errors.Is(BaseException(), None) == False
    with raises(TypeError): errors.Is(1, None)
    with raises(TypeError): errors.Is(object(), None)


    ewrap = Ec(["hello", "world", "мир"])

    assert errors.Is(ewrap, E("мир"))   == True
    assert errors.Is(ewrap, E("май"))   == False

    assert errors.Is(ewrap, Ec(["world", "мир"]))   == True
    assert errors.Is(ewrap, Ec(["hello", "мир"]))   == False
    assert errors.Is(ewrap, Ec(["hello", "май"]))   == False
    assert errors.Is(ewrap, Ec(["world", "май"]))   == False

    assert errors.Is(ewrap, Ec(["hello", "world", "мир"]))  == True
    assert errors.Is(ewrap, Ec(["a",     "world", "мир"]))  == False
    assert errors.Is(ewrap, Ec(["hello", "b",     "мир"]))  == False
    assert errors.Is(ewrap, Ec(["hello", "world", "c"]))    == False

    assert errors.Is(ewrap, Ec(["x", "hello", "world", "мир"])) == False

    # test with XWrap that defines .Unwrap() but not .Error()
    ewrap = XWrap("qqq", Ec(["abc", "привет"]))
    assert errors.Is(ewrap, E("zzz"))               == False
    assert errors.Is(ewrap, E("привет"))            == True
    assert errors.Is(ewrap, Ec(["abc", "привет"]))  == True
    assert errors.Is(ewrap, ewrap)                  == True
    w2 = XWrap("qqq", Ec(["abc", "def"]))
    assert errors.Is(ewrap, w2)                     == False
    w2 = XWrap("qqq", Ec(["abc", "привет"]))
    assert errors.Is(ewrap, w2)                     == True


# ---- Unwrap/Is in the presence of both .Unwrap and .__cause__ ----


# U returns [] with e error chain built via e.Unwrap and e.__cause__ recursion.
def U(e):
    if e is None:
        return []
    return [e] + U(getattr(e,'Unwrap',lambda:None)()) + U(getattr(e,'__cause__',None))

# Uunwrap returns [] with e error chain built via errors.Unwrap recursion.
def Uunwrap(e):
    return [e] + list(_errors._pyUnwrapIter(e))

# verifyUnwrap verifies errors.UnwrapIter via comparing its result with direct
# recursion through .Unwrap and .__cause__ .
def verifyUnwrap(e, estrvok):
    assert [str(_) for _ in U(e)] == estrvok
    assert Uunwrap(e) == U(e)

# verify how errors.Unwrap and errors.Is handle unwrapping when .__cause__ is also !None.
def test_unwrap_with_cause():
    E  = errors.New
    Ec = error_mkchain
    V  = verifyUnwrap

    V(E("abc"), ["abc"])
    V(E("hello world"), ["hello world"])

    #   e1 w→ e2 w→ e3                  e1 w→ e2 w→ e3
    #   |                     ->        |            w
    #   v                               v            |
    #   X                               X ← ← ← ← ← ←
    e1 = Ec(["1", "2", "3"])
    x  = XExc("x")
    e1.__cause__ = x
    V(e1, ["1: 2: 3", "2: 3", "3", "x"])
    assert errors.Is(e1, Ec(["1", "2", "3"]))   == True
    assert errors.Is(e1, Ec(["2", "3"]))        == True
    assert errors.Is(e1, Ec(["2", "4"]))        == False
    assert errors.Is(e1, E("3"))                == True
    assert errors.Is(e1, XExc("x"))             == True
    assert errors.Is(e1, XExc("y"))             == False

    #   e11 w→ e12 w→ e13               e11 w→ e12 w→ e13
    #   |                               | ← ← ← ← ← ← ← w
    #   v                     ->        v↓
    #   e21 w→ e22                      e21 w→ e22
    e11 = Ec(["11", "12", "13"])
    e21 = Ec(["21", "22"])
    e11.__cause__ = e21
    V(e11, ["11: 12: 13", "12: 13", "13", "21: 22", "22"])
    assert errors.Is(e11, Ec(["11", "12", "13"]))   == True
    assert errors.Is(e11, Ec(["12", "13"]))         == True
    assert errors.Is(e11, Ec(["12", "14"]))         == False
    assert errors.Is(e11, E("13"))                  == True
    assert errors.Is(e11, Ec(["11", "12", "13", "21", "22"]))   == False
    assert errors.Is(e11, Ec(["11", "12", "13", "21"]))         == False
    assert errors.Is(e11, Ec(["21", "22"]))         == True
    assert errors.Is(e11, E("22"))                  == True
    assert errors.Is(e11, Ec(["21", "22", "23"]))   == False
    assert errors.Is(e11, Ec(["21", "23"]))         == False

    #   e1 w→ e2 w→ e3
    #   |     |      w
    #   v     v      |
    #   X1← ← X2← ← ←
    e2 = Ec(["2", "3"])
    e1 = EErrorWrap("1", e2)
    x1 = XExc("x1")
    x2 = XExc("x2")
    e1.__cause__ = x1
    e2.__cause__ = x2
    V(e1, ["1: 2: 3", "2: 3", "3", "x2", "x1"])
    assert errors.Is(e1, EErrorWrap("1", Ec(["2", "3"])))   == True
    assert errors.Is(e1, EErrorWrap("1", Ec(["2", "4"])))   == False
    assert errors.Is(e1, EErrorWrap("0", Ec(["2", "3"])))   == False
    assert errors.Is(e1, Ec(["2", "3"]))                    == True
    assert errors.Is(e1, Ec(["2", "4"]))                    == False
    assert errors.Is(e1, XExc("x1"))                        == True
    assert errors.Is(e1, XExc("x2"))                        == True
    assert errors.Is(e1, XExc("x3"))                        == False

    #   e11 w→ e12 w→ e13
    #   |      |       w
    #   v      v .← ← ←'
    #   X1    e21 w→ e22
    #     `← ← ← ← ← ←'
    e13 = XExc("13")
    e12 = EErrorWrap("12", e13)
    e11 = XWrap("11", e12)
    x1  = XExc("x1")
    e21 = Ec(["21", "22"])
    e11.__cause__ = x1
    e12.__cause__ = e21
    V(e11, ["11: 12: 13", "12: 13", "13", "21: 22", "22", "x1"])
    assert errors.Is(e11, EErrorWrap("12", XExc("13")))     == True
    assert errors.Is(e11, EErrorWrap("12", XExc("14")))     == False
    assert errors.Is(e11, EErrorWrap("xx", XExc("13")))     == False
    assert errors.Is(e11, XExc("13"))                       == True
    assert errors.Is(e11, XExc("x1"))                       == True
    assert errors.Is(e11, XExc("y"))                        == False
    assert errors.Is(e11, Ec(["21", "22"]))                 == True
    assert errors.Is(e11, E("22"))                          == True
    assert errors.Is(e11, Ec(["21", "23"]))                 == False
    assert errors.Is(e11, E("23"))                          == False

    #   X1 w→ X2
    #   |     w
    #   v .← ←'
    #   e1 w→ e2 w→ e3
    x2 = XExc("x2")
    x1 = XWrap("x1", x2)
    e1 = Ec(["1", "2", "3"])
    x1.__cause__ = e1
    V(x1, ["x1: x2", "x2", "1: 2: 3", "2: 3", "3"])
    assert errors.Is(x1, XExc("x2"))                == True
    assert errors.Is(x1, XExc("x3"))                == False
    assert errors.Is(x1, XWrap("x1", XExc("x2")))   == True
    assert errors.Is(x1, XWrap("x1", XExc("x3")))   == False
    assert errors.Is(x1, XWrap("y1", XExc("x2")))   == False
    assert errors.Is(x1, Ec(["1", "2", "3"]))       == True
    assert errors.Is(x1, Ec(["2", "3"]))            == True
    assert errors.Is(x1, E("3"))                    == True
    assert errors.Is(x1, Ec(["2", "4"]))            == False

    #   X11 w→ X12
    #   |
    #   v
    #   X21 w→ X22
    x12 = XExc("x12")
    x11 = XWrap("x11", x12)
    x22 = XExc("x22")
    x21 = XWrap("x21", x22)
    x11.__cause__ = x21
    V(x11, ["x11: x12", "x12", "x21: x22", "x22"])
    assert errors.Is(x11, XExc("x12"))                  == True
    assert errors.Is(x11, XExc("x13"))                  == False
    assert errors.Is(x11, XWrap("x11", XExc("x12")))    == True
    assert errors.Is(x11, XWrap("y11", XExc("x12")))    == False
    assert errors.Is(x11, XExc("x22"))                  == True
    assert errors.Is(x11, XExc("x23"))                  == False
    assert errors.Is(x11, XWrap("x21", XExc("x22")))    == True
    assert errors.Is(x11, XWrap("y21", XExc("x22")))    == False
    assert errors.Is(x11, XWrap("x11", XWrap("x21", XExc("x22"))))  == False



# -*- coding: utf-8 -*-
# Copyright (C) 2019-2020  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package sync mirrors and amends Go package sync.

 - `WorkGroup` allows to spawn group of goroutines working on a common task(*).
 - `Once` allows to execute an action only once.
 - `WaitGroup` allows to wait for a collection of tasks to finish.
 - `Sema`(*), `Mutex` and `RWMutex` provide low-level synchronization.

See also https://golang.org/pkg/sync for Go sync package documentation.

(*) not provided in Go standard library, but package
    https://godoc.org/lab.nexedi.com/kirr/go123/xsync
    provides corresponding Go equivalents.
"""

from __future__ import print_function, absolute_import

from golang._sync import \
    PySema      as Sema,        \
    PyMutex     as Mutex,       \
    PyRWMutex   as RWMutex,     \
    PyOnce      as Once,        \
    PyWaitGroup as WaitGroup,   \
    PyWorkGroup as WorkGroup



# -*- coding: utf-8 -*-
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package errors mirrors Go package errors.

 - `New` creates new error with provided text.
 - `Is` tests whether an item in error's chain matches target.

See also https://golang.org/pkg/errors for Go errors package documentation.
See also https://blog.golang.org/go1.13-errors for error chaining overview.
"""

from __future__ import print_function, absolute_import

from golang._errors import \
    pyNew       as New,     \
    pyIs        as Is



# -*- coding: utf-8 -*-
# Copyright (C) 2018-2020  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package strconv provides Go-compatible string conversions."""

from __future__ import print_function, absolute_import

import sys
import six, unicodedata, codecs
from six import text_type as unicode        # py2: unicode      py3: str
from six import unichr                      # py2: unichr       py3: chr
from six import int2byte as bchr            # py2: chr          py3: lambda x: bytes((x,))
from six.moves import range as xrange


# _bstr is like b but also returns whether input was unicode.
def _bstr(s):   # -> sbytes, wasunicode
    wasunicode = False
    if isinstance(s, bytes):                    # py2: str      py3: bytes
        pass
    elif isinstance(s, unicode):                # py2: unicode  py3: str
        wasunicode = True
    else:
        raise TypeError("b: invalid type %s" % type(s))

    if wasunicode:                              # py2: unicode  py3: str
        if six.PY3:
            s = s.encode('UTF-8', 'surrogateescape')
        else:
            # py2 does not have surrogateescape error handler, and even if we
            # provide one, builtin unicode.encode() does not treat
            # \udc80-\udcff as error. -> Do the encoding ourselves.
            s = _utf8_encode_surrogateescape(s)

    return s, wasunicode

# _ustr is like u but also returns whether input was bytes.
def _ustr(s):   # -> sunicode, wasbytes
    wasbytes = True
    if isinstance(s, bytes):                    # py2: str      py3: bytes
        pass
    elif isinstance(s, unicode):                # py2: unicode  py3: str
        wasbytes = False
    else:
        raise TypeError("u: invalid type %s" % type(s))

    if wasbytes:
        if six.PY3:
            s = s.decode('UTF-8', 'surrogateescape')
        else:
            # py2 does not have surrogateescape error handler, and even if we
            # provide one, builtin bytes.decode() does not treat surrogate
            # sequences as error. -> Do the decoding ourselves.
            s = _utf8_decode_surrogateescape(s)

    return s, wasbytes


# quote quotes unicode|bytes string into valid "..." unicode|bytes string always quoted with ".
def quote(s):
    s, wasunicode = _bstr(s)
    qs = _quote(s)
    if wasunicode:
        qs, _ = _ustr(qs)
    return qs

def _quote(s):
    assert isinstance(s, bytes)

    outv = []
    emit = outv.append
    i = 0
    while i < len(s):
        c = s[i:i+1]
        # fast path - ASCII only
        if ord(c) < 0x80:
            if c in b'\\"':
                emit(b'\\'+c)

            # printable ASCII
            elif b' ' <= c <= b'\x7e':
                emit(c)

            # non-printable ASCII
            elif c == b'\t':
                emit(br'\t')
            elif c == b'\n':
                emit(br'\n')
            elif c == b'\r':
                emit(br'\r')

            # everything else is non-printable
            else:
                emit(br'\x%02x' % ord(c))

            i += 1

        # slow path - full UTF-8 decoding + unicodedata
        else:
            r, size = _utf8_decode_rune(s[i:])
            isize = i + size

            # decode error - just emit raw byte as escaped
            if r == _rune_error:
                emit(br'\x%02x' % ord(c))

            # printable utf-8 characters go as is
            elif unicodedata.category(_xunichr(r))[0] in _printable_cat0:
                emit(s[i:isize])

            # everything else goes in numeric byte escapes
            else:
                for j in xrange(i, isize):
                    emit(br'\x%02x' % ord(s[j:j+1]))

            i = isize

    return b'"' + b''.join(outv) + b'"'


# unquote decodes unicode|byte string that was produced by quote.
#
# ValueError is raised if there are quoting syntax errors.
def unquote(s):
    us, tail = unquote_next(s)
    if len(tail) != 0:
        raise ValueError('non-empty tail after closing "')
    return us

# unquote_next decodes next unicode|byte string that was produced by quote.
#
# it returns -> (unquoted(s), tail-after-")
#
# ValueError is raised if there are quoting syntax errors.
def unquote_next(s):
    s, wasunicode = _bstr(s)
    us, tail = _unquote_next(s)
    if wasunicode:
        us, _   = _ustr(us)
        tail, _ = _ustr(tail)
    return us, tail

def _unquote_next(s):
    assert isinstance(s, bytes)

    if len(s) == 0 or s[0:0+1] != b'"':
        raise ValueError('no starting "')

    outv = []
    emit= outv.append

    s = s[1:]
    while 1:
        r, width = _utf8_decode_rune(s)
        if width == 0:
            raise ValueError('no closing "')

        if r == ord('"'):
            s = s[1:]
            break

        # regular UTF-8 character
        if r != ord('\\'):
            emit(s[:width])
            s = s[width:]
            continue

        if len(s) < 2:
            raise ValueError('unexpected EOL after \\')

        c = s[1:1+1]

        # \<c> -> <c>   ; c = \ "
        if c in b'\\"':
            emit(c)
            s = s[2:]
            continue

        if c == b't':
            emit(b'\t')
            s = s[2:]
            continue

        if c == b'n':
            emit(b'\n')
            s = s[2:]
            continue

        if c == b'r':
            emit(b'\r')
            s = s[2:]
            continue

        if c == b'x':   # hex   XXX also handle octals?
            if len(s) < 2+2:
                raise ValueError('unexpected EOL after \\x')

            b = codecs.decode(s[2:2+2], 'hex')
            emit(b)
            s = s[2+2:]
            continue

        raise ValueError('invalid escape \\%s' % chr(ord(c[0:0+1])))

    return b''.join(outv), s


_printable_cat0 = frozenset(['L', 'N', 'P', 'S'])   # letters, numbers, punctuation, symbols

_rune_error = 0xFFFD # unicode replacement character

_ucs2_build        = (sys.maxunicode ==     0xffff)     #    ucs2
assert _ucs2_build or sys.maxunicode >= 0x0010ffff      # or ucs4

# _utf8_decode_rune decodes next UTF8-character from byte string s.
#
# _utf8_decode_rune(s) -> (r, size)
def _utf8_decode_rune(s):
    assert isinstance(s, bytes)

    if len(s) == 0:
        return _rune_error, 0

    l = min(len(s), 4)  # max size of an UTF-8 encoded character
    while l > 0:
        try:
            r = s[:l].decode('utf-8', 'strict')
        except UnicodeDecodeError:
            l -= 1
            continue

        if len(r) == 1:
            return ord(r), l

        # see comment in _utf8_encode_surrogateescape
        if _ucs2_build and len(r) == 2:
            try:
                return _xuniord(r), l
            # e.g. TypeError: ord() expected a character, but string of length 2 found
            except TypeError:
                l -= 1
                continue

        l -= 1
        continue

    # invalid UTF-8
    return _rune_error, 1


# _utf8_decode_surrogateescape mimics s.decode('utf-8', 'surrogateescape') from py3.
def _utf8_decode_surrogateescape(s): # -> unicode
    assert isinstance(s, bytes)
    outv = []
    emit = outv.append

    while len(s) > 0:
        r, width = _utf8_decode_rune(s)
        if r == _rune_error:
            b = ord(s[0])
            assert 0x80 <= b <= 0xff
            emit(unichr(0xdc00 + b))

        # python2 "correctly" decodes surrogates - don't allow that as
        # surrogates are not valid UTF-8:
        # https://github.com/python/cpython/blob/v3.8.1-118-gdbb37aac142/Objects/stringlib/codecs.h#L153-L157
        # (python3 raises UnicodeDecodeError for surrogates)
        elif 0xd800 <= r < 0xdfff:
            for c in s[:width]:
                b = ord(c)
                if c >= 0x80:
                    emit(unichr(0xdc00 + b))
                else:
                    emit(unichr(b))

        else:
            emit(_xunichr(r))

        s = s[width:]

    return u''.join(outv)


# _utf8_encode_surrogateescape mimics s.encode('utf-8', 'surrogateescape') from py3.
def _utf8_encode_surrogateescape(s): # -> bytes
    assert isinstance(s, unicode)
    outv = []
    emit = outv.append

    while len(s) > 0:
        uc = s[0]; s = s[1:]
        c = ord(uc)

        if 0xdc80 <= c <= 0xdcff:
            # surrogate - emit unescaped byte
            emit(bchr(c & 0xff))
            continue

        # in builds with --enable-unicode=ucs2 (default for py2 on macos and windows)
        # python represents unicode points > 0xffff as _two_ unicode characters:
        #
        #   uh = u - 0x10000
        #   c1 = 0xd800 + (uh >> 10)      ; [d800, dbff]
        #   c2 = 0xdc00 + (uh & 0x3ff)    ; [dc00, dfff]
        #
        # if detected - merge those two unicode characters for .encode('utf-8') below
        #
        # this should be only relevant for python2, as python3 switched to "flexible"
        # internal unicode representation: https://www.python.org/dev/peps/pep-0393
        if _ucs2_build and (0xd800 <= c <= 0xdbff):
            if len(s) > 0:
                uc2 = s[0]
                c2 = ord(uc2)
                if 0xdc00 <= c2 <= 0xdfff:
                    uc = uc + uc2
                    s = s[1:]

        emit(uc.encode('utf-8', 'strict'))

    return b''.join(outv)


# _xuniord returns ordinal for a unicode character u.
#
# it works correctly even if u is represented as 2 unicode surrogate points on
# ucs2 python build.
if not _ucs2_build:
    _xuniord = ord
else:
    def _xuniord(u):
        assert isinstance(u, unicode)
        if len(u) == 1:
            return ord(u)

        # see _utf8_encode_surrogateescape for details
        if len(u) == 2:
            c1 = ord(u[0])
            c2 = ord(u[1])
            if (0xd800 <= c1 <= 0xdbff) and (0xdc00 <= c2 <= 0xdfff):
                return 0x10000 | ((c1 - 0xd800) << 10) | (c2 - 0xdc00)

        # let it crash
        return ord(u)


# _xunichr returns unicode character for an ordinal i.
#
# it works correctly even on ucs2 python builds, where ordinals >= 0x10000 are
# represented as 2 unicode pointe.
if not _ucs2_build:
    _xunichr = unichr
else:
    def _xunichr(i):
        if i < 0x10000:
            return unichr(i)

        # see _utf8_encode_surrogateescape for details
        uh = i - 0x10000
        return unichr(0xd800 + (uh >> 10)) + \
               unichr(0xdc00 + (uh & 0x3ff))



# -*- coding: utf-8 -*-
# cython: language_level=2
# cython: c_string_type=str, c_string_encoding=utf8
# distutils: language=c++
#
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""_fmt.pyx implements fmt.pyx - see _fmt.pxd for package overview."""

from __future__ import print_function, absolute_import

from golang cimport pyerror
from golang cimport errors, fmt

# _PyWrapError is the pyerror created by pyErrorf("...: %w", pyerr), for case
# when pyerr is general python error - instead of being just pyerror wrapper
# around raw C-level error.
cdef class _PyWrapError(pyerror):
    cdef str    _prefix
    cdef object _errSuffix

    def __cinit__(_PyWrapError pywerr, str prefix, object errSuffix):
        pywerr._prefix = prefix
        pywerr._errSuffix = errSuffix

    def Unwrap(_PyWrapError pywerr): # -> error | None
        return pywerr._errSuffix
    def Error(_PyWrapError pywerr): # -> str
        esuff = pywerr._errSuffix
        if esuff is None:
            esuff = "%!w(<None>)"   # mimic go
        return "%s: %s" % (pywerr._prefix, esuff)


def pyErrorf(str format, *argv): # -> error
    """Errorf formats text into error.

       format suffix ": %w" is additionally handled as in Go with
       `Errorf("... : %w", ..., err)` creating error that can be unwrapped back to err.
    """
    xpyerr = None
    withW  = False
    if format.endswith(": %w"):
        withW = True
        format = format[:-4]
        xpyerr = argv[-1]
        argv   = argv[:-1]

    # NOTE: this will give TypeError  if format vs args is not right.
    # NOTE: this will give ValueError if %w is used inside suffix-stripped format.
    prefix = format % argv

    if not withW:
        return pyerror.from_error(errors.New(prefix))

    if not (isinstance(xpyerr, BaseException) or xpyerr is None):
        raise TypeError("fmt.Errorf: lastarg to wrap is not error: type(argv[-1])=%r" % type(xpyerr))

    # xpyerr is arbitrary exception class - not a wrapper around C-level error object
    if type(xpyerr) is not pyerror:
        return _PyWrapError(prefix, xpyerr)

    # xpyerr is wrapper around C-level error object
    cdef pyerror pyerr = xpyerr
    return pyerror.from_error(fmt.errorf("%s: %w", <const char *>prefix, pyerr.err))



# -*- coding: utf-8 -*-
# cython: language_level=2
# cython: c_string_type=str, c_string_encoding=utf8
# distutils: language=c++
#
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang cimport error, pyerror, nil, topyexc
from golang cimport errors, fmt


# pyerror_mkchain creates error chain from [] of text.
def pyerror_mkchain(textv):
    cdef error err
    cdef const char *s
    for text in reversed(textv):
        if err == nil:
            err = errors.New(text)
        else:
            s = text
            err = fmt.errorf("%s: %w", s, err)
    return pyerror.from_error(err)


# errors_test.cpp
cdef extern from * nogil:
    """
    extern void _test_errors_new_cpp();
    extern void _test_errors_unwrap_cpp();
    extern void _test_errors_is_cpp();
    """
    void _test_errors_new_cpp()                 except +topyexc
    void _test_errors_unwrap_cpp()              except +topyexc
    void _test_errors_is_cpp()                  except +topyexc

def test_errors_new_cpp():
    with nogil:
        _test_errors_new_cpp()
def test_errors_unwrap_cpp():
    with nogil:
        _test_errors_unwrap_cpp()
def test_errors_is_cpp():
    with nogil:
        _test_errors_is_cpp()



# -*- coding: utf-8 -*-
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang.golang_test import import_pyx_tests

import_pyx_tests("golang._cxx_test")



# -*- coding: utf-8 -*-
# Copyright (C) 2019-2020  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang import error
from golang import errors, fmt, _fmt
from golang.golang_test import import_pyx_tests
from golang.errors_test import Unwrap1
from pytest import raises

import_pyx_tests("golang._fmt_test")


# verify fmt.Errorf with focus on %w (error chaining).
# the rest of formatting is served by built-in python %.
def test_errorf():
    e = fmt.Errorf("abc")
    assert type(e) is error
    assert e.Error()        == "abc"
    assert Unwrap1(e)       is None

    e = fmt.Errorf("hello %d world %s", 123, "мир")
    assert type(e) is error
    assert e.Error()        == "hello 123 world мир"
    assert Unwrap1(e)       is None

    # %w with !exception
    with raises(TypeError): fmt.Errorf(": %w", 1)
    with raises(TypeError): fmt.Errorf(": %w", object())

    # errorf with chaining
    e = errors.New("problem")
    w = fmt.Errorf("%s %s: %w", "op", "file", e)
    assert type(w) is error
    assert w.Error()        == "op file: problem"
    assert Unwrap1(w)       == e
    assert errors.Is(w, e)  == True

    w = fmt.Errorf("%s %s: %s", "op", "file", e)
    assert type(w) is error
    assert w.Error()        == "op file: problem"
    assert Unwrap1(w)       is None
    assert errors.Is(w, e)  == False

    # chaining to !error
    e = RuntimeError("abc")
    w = fmt.Errorf("zzz: %w", e)
    assert type(w) is _fmt._PyWrapError
    assert w.Error()        == "zzz: abc"
    assert Unwrap1(w)       is e    # NOTE is
    assert errors.Is(w, e)  == True
    assert Unwrap1(e)       is None

    # chaining to !error with .Unwrap
    class MyError(Exception):
        def __init__(myerr, op, path, err):
            super(MyError, myerr).__init__(op, path, err)
            myerr.op    = op
            myerr.path  = path
            myerr.err   = err

        def Unwrap(myerr):  return myerr.err
        def __str__(myerr): return "myerror %s %s: %s" % (myerr.op, myerr.path, myerr.err)
        # NOTE: no .Error provided

    e1 = KeyError("not found")
    e  = MyError("load", "zzz", e1)
    w  = fmt.Errorf("yyy: %w", e)
    assert type(w) is _fmt._PyWrapError
    with raises(AttributeError): e.Error
    assert str(e)           == "myerror load zzz: 'not found'"
    assert w.Error()        == "yyy: myerror load zzz: 'not found'"
    assert str(w)           == "yyy: myerror load zzz: 'not found'"
    assert Unwrap1(w)       is e    # NOTE is
    assert Unwrap1(e)       is e1   # NOTE is
    assert Unwrap1(e1)      is None
    assert errors.Is(w, e)  == True
    assert errors.Is(w, e1) == True
    assert errors.Is(e, e1) == True

    # %w with nil error
    e = fmt.Errorf("aaa: %w", None)
    assert type(e) is _fmt._PyWrapError
    assert e.Error()        == "aaa: %!w(<None>)"
    assert str(e)           == "aaa: %!w(<None>)"
    assert Unwrap1(e)       is None

    # multiple %w or ": %w" not as suffix -> ValueError
    with raises(ValueError): fmt.Errorf("%w", e)
    with raises(ValueError): fmt.Errorf(":%w", e)
    with raises(ValueError): fmt.Errorf("a %w : %w", e, e)
    with raises(ValueError): fmt.Errorf("%w hi", e)



# -*- coding: utf-8 -*-
# cython: language_level=2
# distutils: language=c++
#
# Copyright (C) 2019-2020  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang cimport topyexc

# strings_test.cpp
cdef extern from * nogil:
    """
    extern void _test_strings_has_prefix();
    extern void _test_strings_trim_prefix();
    extern void _test_strings_has_suffix();
    extern void _test_strings_trim_suffix();
    extern void _test_strings_split();
    """
    void _test_strings_has_prefix()             except +topyexc
    void _test_strings_trim_prefix()            except +topyexc
    void _test_strings_has_suffix()             except +topyexc
    void _test_strings_trim_suffix()            except +topyexc
    void _test_strings_split()                  except +topyexc
def test_strings_has_prefix():
    with nogil:
        _test_strings_has_prefix()
def test_strings_trim_prefix():
    with nogil:
        _test_strings_trim_prefix()
def test_strings_has_suffix():
    with nogil:
        _test_strings_has_suffix()
def test_strings_trim_suffix():
    with nogil:
        _test_strings_trim_suffix()
def test_strings_split():
    with nogil:
        _test_strings_split()



# cython: language_level=2
# Copyright (C) 2019-2020  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package errors mirrors Go package errors.

 - `New` creates new error with provided text.
 - `Unwrap` tries to extract wrapped error.
 - `Is` tests whether an item in error's chain matches target.

See also https://golang.org/pkg/errors for Go errors package documentation.
See also https://blog.golang.org/go1.13-errors for error chaining overview.
"""

from golang cimport error, string, cbool

cdef extern from "golang/errors.h" namespace "golang::errors" nogil:
    error New(const string& text)
    error Unwrap(error err)
    cbool Is(error err, error target)



# -*- coding: utf-8 -*-
# cython: language_level=2
# cython: c_string_type=str, c_string_encoding=utf8
# distutils: language=c++
#
# Copyright (C) 2020  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""_errors.pyx implements errors.pyx - see _errors.pxd for package overview."""

from __future__ import print_function, absolute_import

from golang cimport pyerror, nil, topyexc
from golang import b as pyb
from golang cimport errors

# Taking both .Unwrap() and .__cause__ into account
#
# Contrary to Go and Pyx/C++ cases, in Python errors could link to each other
# by both .Unwrap() and .__cause__ . Below we show that both error links have
# to be taken into account when building an error's chain, and how:
#
# Consider the following cases
#
#     1. X -> X
#     2. X -> error
#     3. error -> X
#     4. error -> error
#
# where
#
#     "->" means link via .__cause__,
#     X     - exception that does not provide .Unwrap()
#     error - exception that provides .Unwrap()
#
# 1: Since the cause is explicit we want errors.Unwrap to follow "->" link,
#    so that e.g. errors.Is works correctly.
#
# 2: The same.
#
# 3: Consider e.g.
#
#     e1 w→ e2 w→ e3                  w→ denotes what .Unwrap() returns
#     |                               -> denotes .__cause__
#     v
#     X
#
#    this picture is a result of
#
#     try:
#         # call something that raises regular exc X
#     except X:
#         err = dosmth_pygo()
#         raise err from x
#
#    due to the logic in code we a) want to keep X in unwrap sequence of err
#    (it was explicitly specified as cause), and b) we want X to be in the
#    end of unwrap sequence of err, because due to the logic it is the most
#    inner cause:
#
#     e1 w→ e2 w→ e3
#     |            w
#     v            |
#     X ← ← ← ← ← ←
#
# 4: Consider e.g.
#
#     e11 w→ e12 w→ e13
#     |
#     v
#     e21 w→ e22
#
#    Similarly to 3 we want err2 to be kept in unwrap sequence of err1 and to
#    be appended there into tail:
#
#     e11 w→ e12 w→ e13
#     | ← ← ← ← ← ← ← w
#     v↓
#     e21 w→ e22
#
# 1-4 suggest the following rules: let U(e) be the list with full error chain
# of e, starting from e itself. For example for `err = e1 w→ e2 w→ e3`,
# `U(err) = [e1, e2, e3]`
#
# we can deduce recursive formula for U based on 1-4:
#
# - U(nil) = []                                           ; obvious
#
# - for e:                U(e) = [e]                      ; obvious
#     .w = nil
#     .__cause__ = nil
#
# - for e:                U(e) = [e] + U(e.w)             ; regular go-style wrapping
#     .w != nil
#     .__cause__ = nil
#
# - for e:                U(e) = [e] + U(e.__cause__)     ; cases 1 & 2
#     .w = nil
#     .__cause__ != nil
#
# - for e:                U(e) = [e] + U(e.w) + U(e.__cause__)    ; ex. cases 3 & 4
#     .w != nil
#     .__cause__ != nil
#
# the formula for cases 3 & 4 is the general one and works for all cases:
#
#     U(nil) = []
#     U(e)   = [e] + U(e.w) + U(e.__cause__)
#
#
# e.g. consider how it works for:
#
#     e1 w→ e2 w→ e3
#     |     |      w
#     v     v      |
#     X1← ← X2← ← ←
#
# U(e1) = [e1] + U(e2) + U(X1)
#       = [e1] + {[e2] + U(e3) + U(X2)} + [X1]
#       = [e1] + {[e2] + [e3] + [X2]} + [X1]
#       = [e1, e2, e3, X2, X1]
#
# --------
#
# Implementation: with errors.Unwrap we cannot use U directly to implement
# unwrapping because it requires to keep unwrapping iterator state and
# errors.Unwrap API returns just an error instance, nothing more. For this
# reason python version of errors package, does not expose errors.Unwrap, and
# internally uses errors._UnwrapIter, which returns iterator through an
# error's error chain.


def pyNew(text): # -> error
    """New creates new error with provided text."""
    return pyerror.from_error(errors_New_pyexc(pyb(text)))


def _pyUnwrapIter(err): # -> iter(error)
    """_UnwrapIter returns iterator through err's error chain.

       This iteration takes both .Unwrap() and .__cause__ into account.
       See "Taking both .Unwrap() and .__cause__ into account" in internal overview.
    """
    if err is None:
        return
    if not isinstance(err, BaseException):
        raise TypeError("errors.UnwrapIter: err is not exception: type(err)=%r" % type(err))

    cdef pyerror pye
    cdef error   e

    # + U(e.w)
    if type(err) is not pyerror:
        # err is python-level error (pyerror-based or just BaseException child)
        eunwrap = getattr(err, 'Unwrap', _missing)
        pyw = None
        if eunwrap is not _missing:
            pyw = eunwrap()
        if pyw is not None:
            yield pyw
            for _ in _pyUnwrapIter(pyw):
                yield _
    else:
        # err is wrapper around C-level error
        pye = err
        e   = pye.err
        while 1:
            e = errors_Unwrap_pyexc(e)
            if e == nil:
                break
            yield pyerror.from_error(e)

    # + U(e.__cause__)
    pycause    = getattr(err, '__cause__', None)
    if pycause is not None:
        yield pycause
        for _ in _pyUnwrapIter(pycause):
            yield _


def pyIs(err, target): # -> bool
    """Is returns whether target matches any error in err's error chain."""

    # err and target must be exception or None
    if not (isinstance(err, BaseException) or err is None):
        raise TypeError("errors.Is: err is not exception or None: type(err)=%r" % type(err))
    if not (isinstance(target, BaseException) or target is None):
        raise TypeError("errors.Is: target is not exception or None: type(target)=%r" % type(target))

    if target is None:
        return (err is None)

    wit = _pyUnwrapIter(err)
    while 1:
        if err is None:
            return False

        if type(err) is type(target):
            if err == target:
                return True

        err = next(wit, None)


# ---- misc ----

cdef _missing = object()

cdef nogil:

    error errors_New_pyexc(const char* text)            except +topyexc:
        return errors.New(text)

    error errors_Unwrap_pyexc(error err)                except +topyexc:
        return errors.Unwrap(err)
